# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from stanza.hub.v1 import config_pb2 as stanza_dot_hub_dot_v1_dot_config__pb2


class ConfigServiceStub(object):
    """This protocol buffer definition is the source of truth for the Stanza Hub Quota API.
    There is also an OpenAPI spec, generated via https://github.com/grpc-ecosystem/grpc-gateway.

    Config is used by the language-specific Stanza SDKs and Browser SDK to fetch Service and Feature configurations.
    GuardConfig: Guard level configuration ()
    BrowserContext: Whether Feature is enabled or disabled, has centralized ratelimiting, etc
    ServiceConfig: Service level configuration (OTEL defaults, Sentinel configs, etc)

    The Config service definition.
    This service is used by the Stanza SDK to get Service and Guard configs.
    This service is used by the Browser SDK to get Browser Contexts.
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.GetGuardConfig = channel.unary_unary(
                '/stanza.hub.v1.ConfigService/GetGuardConfig',
                request_serializer=stanza_dot_hub_dot_v1_dot_config__pb2.GetGuardConfigRequest.SerializeToString,
                response_deserializer=stanza_dot_hub_dot_v1_dot_config__pb2.GetGuardConfigResponse.FromString,
                )
        self.GetBrowserContext = channel.unary_unary(
                '/stanza.hub.v1.ConfigService/GetBrowserContext',
                request_serializer=stanza_dot_hub_dot_v1_dot_config__pb2.GetBrowserContextRequest.SerializeToString,
                response_deserializer=stanza_dot_hub_dot_v1_dot_config__pb2.GetBrowserContextResponse.FromString,
                )
        self.GetServiceConfig = channel.unary_unary(
                '/stanza.hub.v1.ConfigService/GetServiceConfig',
                request_serializer=stanza_dot_hub_dot_v1_dot_config__pb2.GetServiceConfigRequest.SerializeToString,
                response_deserializer=stanza_dot_hub_dot_v1_dot_config__pb2.GetServiceConfigResponse.FromString,
                )


class ConfigServiceServicer(object):
    """This protocol buffer definition is the source of truth for the Stanza Hub Quota API.
    There is also an OpenAPI spec, generated via https://github.com/grpc-ecosystem/grpc-gateway.

    Config is used by the language-specific Stanza SDKs and Browser SDK to fetch Service and Feature configurations.
    GuardConfig: Guard level configuration ()
    BrowserContext: Whether Feature is enabled or disabled, has centralized ratelimiting, etc
    ServiceConfig: Service level configuration (OTEL defaults, Sentinel configs, etc)

    The Config service definition.
    This service is used by the Stanza SDK to get Service and Guard configs.
    This service is used by the Browser SDK to get Browser Contexts.
    """

    def GetGuardConfig(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetBrowserContext(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetServiceConfig(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_ConfigServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'GetGuardConfig': grpc.unary_unary_rpc_method_handler(
                    servicer.GetGuardConfig,
                    request_deserializer=stanza_dot_hub_dot_v1_dot_config__pb2.GetGuardConfigRequest.FromString,
                    response_serializer=stanza_dot_hub_dot_v1_dot_config__pb2.GetGuardConfigResponse.SerializeToString,
            ),
            'GetBrowserContext': grpc.unary_unary_rpc_method_handler(
                    servicer.GetBrowserContext,
                    request_deserializer=stanza_dot_hub_dot_v1_dot_config__pb2.GetBrowserContextRequest.FromString,
                    response_serializer=stanza_dot_hub_dot_v1_dot_config__pb2.GetBrowserContextResponse.SerializeToString,
            ),
            'GetServiceConfig': grpc.unary_unary_rpc_method_handler(
                    servicer.GetServiceConfig,
                    request_deserializer=stanza_dot_hub_dot_v1_dot_config__pb2.GetServiceConfigRequest.FromString,
                    response_serializer=stanza_dot_hub_dot_v1_dot_config__pb2.GetServiceConfigResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'stanza.hub.v1.ConfigService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class ConfigService(object):
    """This protocol buffer definition is the source of truth for the Stanza Hub Quota API.
    There is also an OpenAPI spec, generated via https://github.com/grpc-ecosystem/grpc-gateway.

    Config is used by the language-specific Stanza SDKs and Browser SDK to fetch Service and Feature configurations.
    GuardConfig: Guard level configuration ()
    BrowserContext: Whether Feature is enabled or disabled, has centralized ratelimiting, etc
    ServiceConfig: Service level configuration (OTEL defaults, Sentinel configs, etc)

    The Config service definition.
    This service is used by the Stanza SDK to get Service and Guard configs.
    This service is used by the Browser SDK to get Browser Contexts.
    """

    @staticmethod
    def GetGuardConfig(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/stanza.hub.v1.ConfigService/GetGuardConfig',
            stanza_dot_hub_dot_v1_dot_config__pb2.GetGuardConfigRequest.SerializeToString,
            stanza_dot_hub_dot_v1_dot_config__pb2.GetGuardConfigResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetBrowserContext(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/stanza.hub.v1.ConfigService/GetBrowserContext',
            stanza_dot_hub_dot_v1_dot_config__pb2.GetBrowserContextRequest.SerializeToString,
            stanza_dot_hub_dot_v1_dot_config__pb2.GetBrowserContextResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetServiceConfig(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/stanza.hub.v1.ConfigService/GetServiceConfig',
            stanza_dot_hub_dot_v1_dot_config__pb2.GetServiceConfigRequest.SerializeToString,
            stanza_dot_hub_dot_v1_dot_config__pb2.GetServiceConfigResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: stanza/hub/v1/usage.proto
# Protobuf Python Version: 4.25.2
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.api import annotations_pb2 as google_dot_api_dot_annotations__pb2
from google.api import field_behavior_pb2 as google_dot_api_dot_field__behavior__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from stanza.hub.v1 import common_pb2 as stanza_dot_hub_dot_v1_dot_common__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x19stanza/hub/v1/usage.proto\x12\rstanza.hub.v1\x1a\x1cgoogle/api/annotations.proto\x1a\x1fgoogle/api/field_behavior.proto\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x1astanza/hub/v1/common.proto\"\xb9\x07\n\x0fGetUsageRequest\x12&\n\x0b\x65nvironment\x18\x01 \x01(\tB\x04\xe2\x41\x01\x02R\x0b\x65nvironment\x12\x19\n\x05guard\x18\x02 \x01(\tH\x00R\x05guard\x88\x01\x01\x12G\n\x10guard_query_mode\x18\x03 \x01(\x0e\x32\x18.stanza.hub.v1.QueryModeH\x01R\x0eguardQueryMode\x88\x01\x01\x12;\n\x08start_ts\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x04\xe2\x41\x01\x02R\x07startTs\x12\x37\n\x06\x65nd_ts\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x04\xe2\x41\x01\x02R\x05\x65ndTs\x12\x1b\n\x06\x61pikey\x18\x06 \x01(\tH\x02R\x06\x61pikey\x88\x01\x01\x12\x1d\n\x07\x66\x65\x61ture\x18\x07 \x01(\tH\x03R\x07\x66\x65\x61ture\x88\x01\x01\x12K\n\x12\x66\x65\x61ture_query_mode\x18\x08 \x01(\x0e\x32\x18.stanza.hub.v1.QueryModeH\x04R\x10\x66\x65\x61tureQueryMode\x88\x01\x01\x12\x1d\n\x07service\x18\x0f \x01(\tH\x05R\x07service\x88\x01\x01\x12K\n\x12service_query_mode\x18\x10 \x01(\x0e\x32\x18.stanza.hub.v1.QueryModeH\x06R\x10serviceQueryMode\x88\x01\x01\x12\x1f\n\x08priority\x18\t \x01(\x05H\x07R\x08priority\x88\x01\x01\x12M\n\x13priority_query_mode\x18\n \x01(\x0e\x32\x18.stanza.hub.v1.QueryModeH\x08R\x11priorityQueryMode\x88\x01\x01\x12\x1f\n\x0breport_tags\x18\x0b \x03(\tR\nreportTags\x12&\n\x04tags\x18\x0c \x03(\x0b\x32\x12.stanza.hub.v1.TagR\x04tags\x12+\n\x0freport_all_tags\x18\r \x01(\x08H\tR\rreportAllTags\x88\x01\x01\x12\x17\n\x04step\x18\x0e \x01(\tH\nR\x04step\x88\x01\x01\x42\x08\n\x06_guardB\x13\n\x11_guard_query_modeB\t\n\x07_apikeyB\n\n\x08_featureB\x15\n\x13_feature_query_modeB\n\n\x08_serviceB\x15\n\x13_service_query_modeB\x0b\n\t_priorityB\x16\n\x14_priority_query_modeB\x12\n\x10_report_all_tagsB\x07\n\x05_step\"J\n\x10GetUsageResponse\x12\x36\n\x06result\x18\x01 \x03(\x0b\x32\x1e.stanza.hub.v1.UsageTimeseriesR\x06result\"\x97\x02\n\x0fUsageTimeseries\x12\x33\n\x04\x64\x61ta\x18\x01 \x03(\x0b\x32\x1f.stanza.hub.v1.UsageTSDataPointR\x04\x64\x61ta\x12\x1d\n\x07\x66\x65\x61ture\x18\x03 \x01(\tH\x00R\x07\x66\x65\x61ture\x88\x01\x01\x12\x1f\n\x08priority\x18\x04 \x01(\x05H\x01R\x08priority\x88\x01\x01\x12&\n\x04tags\x18\x05 \x03(\x0b\x32\x12.stanza.hub.v1.TagR\x04tags\x12\x19\n\x05guard\x18\x06 \x01(\tH\x02R\x05guard\x88\x01\x01\x12\x1d\n\x07service\x18\x07 \x01(\tH\x03R\x07service\x88\x01\x01\x42\n\n\x08_featureB\x0b\n\t_priorityB\x08\n\x06_guardB\n\n\x08_service\"\x86\x04\n\x10UsageTSDataPoint\x12\x35\n\x08start_ts\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x07startTs\x12\x31\n\x06\x65nd_ts\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x05\x65ndTs\x12\x18\n\x07granted\x18\x03 \x01(\x05R\x07granted\x12%\n\x0egranted_weight\x18\x04 \x01(\x02R\rgrantedWeight\x12\x1f\n\x0bnot_granted\x18\x05 \x01(\x05R\nnotGranted\x12,\n\x12not_granted_weight\x18\x06 \x01(\x02R\x10notGrantedWeight\x12\x1e\n\x08\x62\x65_burst\x18\x07 \x01(\x05H\x00R\x07\x62\x65\x42urst\x88\x01\x01\x12+\n\x0f\x62\x65_burst_weight\x18\x08 \x01(\x02H\x01R\rbeBurstWeight\x88\x01\x01\x12(\n\rparent_reject\x18\t \x01(\x05H\x02R\x0cparentReject\x88\x01\x01\x12\x35\n\x14parent_reject_weight\x18\n \x01(\x02H\x03R\x12parentRejectWeight\x88\x01\x01\x42\x0b\n\t_be_burstB\x12\n\x10_be_burst_weightB\x10\n\x0e_parent_rejectB\x17\n\x15_parent_reject_weight*R\n\tQueryMode\x12\x1a\n\x16QUERY_MODE_UNSPECIFIED\x10\x00\x12\x12\n\x0eQUERY_MODE_SUM\x10\x01\x12\x15\n\x11QUERY_MODE_REPORT\x10\x02\x32q\n\x0cUsageService\x12\x61\n\x08GetUsage\x12\x1e.stanza.hub.v1.GetUsageRequest\x1a\x1f.stanza.hub.v1.GetUsageResponse\"\x14\x82\xd3\xe4\x93\x02\x0e\"\t/v1/usage:\x01*B\x8a\x01\n\x11\x63om.stanza.hub.v1B\nUsageProtoP\x01Z\x13stanza/hub/v1;hubpb\xa2\x02\x03SHX\xaa\x02\rStanza.Hub.V1\xca\x02\rStanza\\Hub\\V1\xe2\x02\x19Stanza\\Hub\\V1\\GPBMetadata\xea\x02\x0fStanza::Hub::V1b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'stanza.hub.v1.usage_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\021com.stanza.hub.v1B\nUsageProtoP\001Z\023stanza/hub/v1;hubpb\242\002\003SHX\252\002\rStanza.Hub.V1\312\002\rStanza\\Hub\\V1\342\002\031Stanza\\Hub\\V1\\GPBMetadata\352\002\017Stanza::Hub::V1'
  _globals['_GETUSAGEREQUEST'].fields_by_name['environment']._options = None
  _globals['_GETUSAGEREQUEST'].fields_by_name['environment']._serialized_options = b'\342A\001\002'
  _globals['_GETUSAGEREQUEST'].fields_by_name['start_ts']._options = None
  _globals['_GETUSAGEREQUEST'].fields_by_name['start_ts']._serialized_options = b'\342A\001\002'
  _globals['_GETUSAGEREQUEST'].fields_by_name['end_ts']._options = None
  _globals['_GETUSAGEREQUEST'].fields_by_name['end_ts']._serialized_options = b'\342A\001\002'
  _globals['_USAGESERVICE'].methods_by_name['GetUsage']._options = None
  _globals['_USAGESERVICE'].methods_by_name['GetUsage']._serialized_options = b'\202\323\344\223\002\016\"\t/v1/usage:\001*'
  _globals['_QUERYMODE']._serialized_start=2003
  _globals['_QUERYMODE']._serialized_end=2085
  _globals['_GETUSAGEREQUEST']._serialized_start=169
  _globals['_GETUSAGEREQUEST']._serialized_end=1122
  _globals['_GETUSAGERESPONSE']._serialized_start=1124
  _globals['_GETUSAGERESPONSE']._serialized_end=1198
  _globals['_USAGETIMESERIES']._serialized_start=1201
  _globals['_USAGETIMESERIES']._serialized_end=1480
  _globals['_USAGETSDATAPOINT']._serialized_start=1483
  _globals['_USAGETSDATAPOINT']._serialized_end=2001
  _globals['_USAGESERVICE']._serialized_start=2087
  _globals['_USAGESERVICE']._serialized_end=2200
# @@protoc_insertion_point(module_scope)



import asyncio
import datetime
import logging
import threading
import time
from collections import defaultdict
from datetime import timedelta, timezone
from enum import Enum
from typing import Dict, Iterable, List, Optional, cast

import grpc
from getstanza.hub import StanzaHub
from getstanza.otel import StanzaMeter
from getstanza.propagation import get_feature, get_priority_boost
from opentelemetry.trace import Span, StatusCode
from opentelemetry.util.types import Attributes
from stanza.hub.v1 import common_pb2, config_pb2, quota_pb2
from stanza.hub.v1.common_pb2 import Config, Local, Mode, Quota, Token

# Definition for the 200ms deadline for communicated consumed tokens to Hub.
# Specification Link: https://github.com/StanzaSystems/sdk-spec#token-leases
BATCH_TOKEN_CONSUME_INTERVAL = 0.2

DEFAULT_TIMEOUT = 300
TOKEN_LEASE_TIMEOUT = 200

# Contains all cached leases returned by Hub that haven't been used yet.
cached_leases: defaultdict[
    tuple, List[tuple[datetime.datetime, quota_pb2.TokenLease]]  # Expiration and lease
] = defaultdict(list)
cached_leases_lock = threading.Lock()

# Contains all leases that have been consumed, but not yet communicated to Hub.
consumed_leases: defaultdict[str, List[quota_pb2.TokenLease]] = defaultdict(list)
consumed_leases_lock = threading.Lock()

# An asynchronous task that flushes consumed leases to Hub every ~200ms.
batch_token_consumer_handle: Optional[asyncio.Handle] = None
batch_token_consumer_handle_lock = threading.Lock()


class GuardedStatus(Enum):
    """Indicate success or failure of the guarded code block."""

    GUARDED_UNKNOWN = 0
    GUARDED_SUCCESS = 1
    GUARDED_FAILURE = 2


class GuardEvent(Enum):
    """Possible event types a guard may emit."""

    ALLOWED = 0
    BLOCKED = 1
    FAILOPEN = 2


class TokenLeaseStatus(Enum):
    """Represents the possible statuses of a cached token lease."""

    TOKEN_LEASE_VALID = 0
    TOKEN_LEASE_EXPIRED = 1
    TOKEN_LEASE_WRONG_PRIORITY = 2


class Guard:
    """
    Generic Guard class that implements logic for interacting with Stanza
    Guards. All framework-specific logic is kept out of this implementation.
    """

    @property
    def success(self):
        return GuardedStatus.GUARDED_SUCCESS

    @property
    def failure(self):
        return GuardedStatus.GUARDED_FAILURE

    @property
    def config_status(self):
        return self.__config_status

    @property
    def local_status(self):
        return self.__local_status

    @property
    def token_status(self):
        return self.__token_status

    @property
    def quota_status(self):
        return self.__quota_status

    @property
    def quota_token(self):
        return self.__quota_token

    @property
    def guard_config(self):
        return self.__guard_config

    @property
    def error(self):
        """If there was an error, return the message as a string."""

        return self.__error_message

    @property
    def block_message(self) -> Optional[str]:
        """Returns the reason for the block as a human readable string."""

        if self.__local_status is Local.LOCAL_BLOCKED:
            return "Local resource exhausted. Please try again later."

        if self.__token_status is Token.TOKEN_NOT_VALID:
            return "Invalid or expired X-Stanza-Token."

        if self.__quota_status is Quota.QUOTA_BLOCKED:
            return "Stanza quota exhausted. Please try again later."

        return None

    @property
    def block_reason(self) -> Optional[str]:
        """Returns the reason for the block as a string."""

        if self.__local_status is Local.LOCAL_BLOCKED:
            return Local.Name(self.__local_status)

        if self.__token_status is Token.TOKEN_NOT_VALID:
            return Token.Name(self.__token_status)

        if self.__quota_status is Quota.QUOTA_BLOCKED:
            return Quota.Name(self.__quota_status)

        return None

    @property
    def __cached_lease_key(self):
        """Returns the cache key needed to find cached leases for the guard."""

        return (
            str(self.__client_config.environment or ""),
            self.__guard_name,
            self.__feature_name,
        )

    @property
    def __consumed_lease_key(self):
        """Returns the key needed to group consumed leases by env."""

        return str(self.__client_config.environment or "")

    def __init__(
        self,
        hub: StanzaHub,
        guard_name: str,
        feature_name: Optional[str] = None,
        priority_boost: Optional[int] = None,
        default_weight: Optional[float] = None,
        tags: Optional[Dict[str, str]] = None,
    ):
        global batch_token_consumer_handle

        self.__guard_name = guard_name
        self.__feature_name = get_feature(feature_name)
        self.__priority_boost = get_priority_boost(priority_boost) or 0
        self.__default_weight = 0 if default_weight is None else default_weight
        self.__tags = tags

        self.__event_loop = hub.event_loop
        self.__quota_service = hub.quota_service
        self.__config_manager = hub.config_manager
        self.__client_config = hub.config_manager.config

        self.__meter: Optional[StanzaMeter] = None
        self.__span: Optional[Span] = None
        self.__start: Optional[datetime.datetime] = None
        self.__quota_token: Optional[str] = None
        self.__error_message: Optional[str] = None
        self.__guard_config: Optional[config_pb2.GuardConfig] = None

        self.__mode = Mode.MODE_NORMAL
        self.__config_status = Config.CONFIG_UNSPECIFIED
        self.__local_status = Local.LOCAL_NOT_EVAL
        self.__token_status = Token.TOKEN_NOT_EVAL
        self.__quota_status = Quota.QUOTA_NOT_EVAL

        # Start the task to flush consumed token leases if not already started.
        with batch_token_consumer_handle_lock:
            if batch_token_consumer_handle is None:
                batch_token_consumer_handle = self.__event_loop.call_soon_threadsafe(
                    start_batch_token_consumer(self.__event_loop)
                )

    def __repr__(self) -> str:
        """Returns all of the current status state of the guard."""

        return "guard={}, config_state={}, local_reason={}, token_reason={}, quota_reason={}".format(
            self.__guard_name,
            Config.Name(self.__config_status),
            Local.Name(self.__local_status),
            Token.Name(self.__token_status),
            Quota.Name(self.__quota_status),
        )

    async def run(self, tokens: Optional[Iterable[str]] = None):
        """Run all guard checks and update guard statuses."""

        if self.__config_manager.otel:
            if self.__config_manager.otel.meter:
                self.__meter = self.__config_manager.otel.meter
            if self.__config_manager.otel.tracer:
                self.__span = self.__config_manager.otel.tracer.start_span(
                    "stanza-guard"
                )

        try:
            # Config state check
            if not await self.__check_config():
                return

            # Local rules check
            if not self.__check_local():
                return

            # Ingress token check
            if not self.__check_token(tokens):
                return

            # Quota check
            self.__check_quota()

        except Exception as exc:
            if self.__span:
                self.__span.record_exception(exc)

            if not self.__error_message:
                template = "an unknown exception of type {0} occurred"
                self.__error_message = template.format(type(exc).__name__)

        finally:
            if self.__error_message:
                self.__failopen()
            elif self.allowed():
                self.__allowed()
            else:
                self.__blocked()

            if self.__span:
                self.__span.end()

    async def __check_config(self) -> bool:
        """Check guard configuration."""

        if self.__guard_config is None:
            (
                self.__guard_config,
                self.__config_status,
            ) = await self.__config_manager.get_guard_config(self.__guard_name)

        if self.__guard_config is not None:
            if self.__guard_config.report_only:
                self.__mode = Mode.MODE_REPORT_ONLY
            if (
                self.__config_status == Config.CONFIG_CACHED_OK
                or self.__config_status == Config.CONFIG_FETCHED_OK
            ):
                return True

        self.__error_message = "unable to fetch guard config: {}".format(
            Config.Name(self.__config_status)
        )
        return False

    def __check_local(self) -> bool:
        """Local check is not currently supported by this SDK."""

        self.__local_status = Local.LOCAL_NOT_SUPPORTED
        return True

    def __check_token(self, tokens: Optional[Iterable[str]] = None) -> bool:
        """Validate using the ingress token if configured to do so."""

        if not self.__guard_config or not self.__guard_config.validate_ingress_tokens:
            self.__token_status = Token.TOKEN_EVAL_DISABLED
            return True

        if not tokens:
            self.__token_status = Token.TOKEN_NOT_VALID
            return False

        tokens_info = list(
            map(
                lambda token: quota_pb2.TokenInfo(
                    token=token,
                    guard=common_pb2.GuardSelector(
                        environment=self.__client_config.environment,
                        name=self.__guard_name,
                    ),
                ),
                tokens or [],
            )
        )

        try:
            validate_token_response = cast(
                quota_pb2.ValidateTokenResponse,
                self.__quota_service.ValidateToken(
                    request=quota_pb2.ValidateTokenRequest(tokens=tokens_info),
                    metadata=self.__client_config.metadata,
                    timeout=DEFAULT_TIMEOUT,
                ),
            )
        except grpc.RpcError as rpc_error:
            self.__error_message = str(rpc_error)
            self.__token_status = Token.TOKEN_VALIDATION_ERROR
            return False

        if validate_token_response.valid:
            self.__token_status = Token.TOKEN_VALID
            return True
        else:
            self.__token_status = Token.TOKEN_NOT_VALID
            return False

    def __check_quota(self) -> bool:
        """Quota check using token leases."""

        if not self.__guard_config or not self.__guard_config.check_quota:
            self.__quota_status = Quota.QUOTA_EVAL_DISABLED
            return True

        # Check to see if a cached lease can be found using the guard selector.
        if cached_lease := self.__consume_cached_token_lease():
            logging.debug("Found a valid cached token lease: %s", cached_lease.token)
            self.__quota_status = Quota.QUOTA_GRANTED
            self.__quota_token = cached_lease.token
            return True

        # Add tags (if Guard config allows it)
        tags: List[common_pb2.Tag] = []
        if self.__tags and self.__guard_config:
            for key, value in self.__tags.items():
                if key in self.__guard_config.quota_tags:
                    tags.append(common_pb2.Tag(key=key, value=value))
                else:
                    logging.info(
                        "skipping unknown tag = %s, guard = %s", key, self.__guard_name
                    )

        token_lease_request = quota_pb2.GetTokenLeaseRequest(
            selector=common_pb2.GuardFeatureSelector(
                environment=self.__client_config.environment,
                guard_name=self.__guard_name,
                feature_name=self.__feature_name,
                tags=tags,
            ),
            client_id=self.__client_config.client_id,
            priority_boost=self.__priority_boost,
            default_weight=self.__default_weight,
        )

        try:
            logging.debug(
                "Requesting a token lease with selector: ("
                "environment = %s, "
                "guard = %s, "
                "feature = %s, "
                "priority boost = %s, "
                "default weight = %s, "
                "tags = %s)",
                self.__client_config.environment,
                self.__guard_name,
                self.__feature_name,
                self.__priority_boost,
                self.__default_weight,
                self.__tags,
            )

            token_lease_response = cast(
                quota_pb2.GetTokenLeaseResponse,
                self.__quota_service.GetTokenLease(
                    request=token_lease_request,
                    metadata=self.__client_config.metadata,
                    timeout=TOKEN_LEASE_TIMEOUT,
                ),
            )
        except grpc.RpcError as rpc_error:
            self.__error_message = str(rpc_error)
            self.__quota_status = Quota.QUOTA_ERROR
            return False

        # Immediately consume a lease upon being granted token leases, and
        # cache the rest for later use; otherwise set quota status to BLOCKED.
        if token_lease_response.granted:
            lease = token_lease_response.leases[0]
            self.__consume_token_lease(lease)

            logging.debug(
                "Received %d token leases: %s",
                len(token_lease_response.leases),
                token_lease_response.leases,
            )

            if len(token_lease_response.leases) > 1:
                self.__set_cached_token_leases(token_lease_response.leases[1:])

            self.__quota_status = Quota.QUOTA_GRANTED
            self.__quota_token = lease.token
            return True

        logging.debug("Token lease request has been denied")
        self.__quota_status = Quota.QUOTA_BLOCKED
        return False

    def allowed(self) -> bool:
        """Check if the Guard is currently allowing traffic."""

        # Allow if config is unspecified or failed to fetch the first time.
        if self.__guard_config is None and self.__config_status in [
            Config.CONFIG_UNSPECIFIED,
            Config.CONFIG_FETCH_ERROR,
            Config.CONFIG_FETCH_TIMEOUT,
        ]:
            return True

        # Always allow traffic when 'report only' is set.
        if self.__mode == Mode.MODE_REPORT_ONLY:
            return True

        # Allow if all of the following checks have succeeded.
        return (
            (
                self.__config_status == Config.CONFIG_CACHED_OK
                or self.__config_status == Config.CONFIG_FETCHED_OK
            )
            and self.__local_status != Local.LOCAL_BLOCKED
            and self.__quota_status != Quota.QUOTA_BLOCKED
            and self.__token_status != Token.TOKEN_NOT_VALID
        )

    def blocked(self) -> bool:
        """Check if the Guard is currently disallowing traffic."""

        return not self.allowed()

    def end(self, status: GuardedStatus):
        """Called when the guarded logic comes to an end."""
        if self.__meter:
            if self.__start:
                duration_ms = (
                    datetime.datetime.now() - self.__start
                ) / datetime.timedelta(milliseconds=1)
                self.__meter.AllowedDuration.record(duration_ms, self.__attributes())
            if status == GuardedStatus.GUARDED_SUCCESS:
                self.__meter.AllowedSuccessCount.add(1, self.__attributes())
            elif status == GuardedStatus.GUARDED_FAILURE:
                self.__meter.AllowedFailureCount.add(1, self.__attributes())
            else:
                self.__meter.AllowedUnknownCount.add(1, self.__attributes())

    def __attributes(self) -> Attributes:
        attr: Attributes = {
            "client_id": self.__client_config.client_id,
            "customer_id": str(self.__client_config.customer_id or ""),
            "environment": str(self.__client_config.environment or ""),
            "guard": self.__guard_name,
            "feature": str(self.__feature_name or ""),
            "service": str(self.__client_config.service_name or ""),
            "priority_boost": int(self.__priority_boost),
            "mode": Mode.Name(self.__mode),
            "config_state": Config.Name(self.__config_status),
            "local_reason": Local.Name(self.__local_status),
            "token_reason": Token.Name(self.__token_status),
            "quota_reason": Quota.Name(self.__quota_status),
        }
        return attr

    def __emit_event(self, guard_event: GuardEvent, message: str):
        """Log any type of guard event."""

        if message != "":
            logging.debug(f"{message}, %s", repr(self))

        if self.__meter:
            if guard_event == GuardEvent.ALLOWED:
                self.__meter.AllowedCount.add(1, self.__attributes())
            elif guard_event == GuardEvent.BLOCKED:
                self.__meter.BlockedCount.add(1, self.__attributes())
            elif guard_event == GuardEvent.FAILOPEN:
                self.__meter.FailOpenCount.add(1, self.__attributes())

        if self.__span:
            if guard_event == GuardEvent.ALLOWED:
                self.__span.add_event("Stanza allowed", self.__attributes())
                self.__span.set_status(StatusCode.OK)
            elif guard_event == GuardEvent.BLOCKED:
                self.__span.add_event("Stanza blocked", self.__attributes())
                self.__span.set_status(StatusCode.ERROR, message)
            elif guard_event == GuardEvent.FAILOPEN:
                self.__span.add_event("Stanza failed open", self.__attributes())
                self.__span.set_status(StatusCode.ERROR, message)

    def __allowed(self):
        """Log an allowed event."""

        self.__emit_event(GuardEvent.ALLOWED, "Stanza allowed")
        self.__start = datetime.datetime.now()

    def __blocked(self):
        """Log a blocked event."""

        self.__emit_event(GuardEvent.BLOCKED, "Stanza blocked")

    def __failopen(self):
        """Log a failopen event."""

        self.__emit_event(
            GuardEvent.FAILOPEN, f"Stanza failed open, {self.__error_message}"
        )

    def __consume_cached_token_lease(self) -> Optional[quota_pb2.TokenLease]:
        """Scans cached leases and finds the first valid and unexpired lease."""

        with cached_leases_lock:
            if self.__cached_lease_key not in cached_leases:
                return None

            num_discarded_leases = 0

            # Scan through leases until a valid and unexpired one is found. Any
            # invalid leases found during this process are discarded.
            for i, cached_lease in enumerate(cached_leases[self.__cached_lease_key][:]):
                expiration, lease = cached_lease
                token_status = self.__check_token_lease(lease, expiration)

                if token_status == TokenLeaseStatus.TOKEN_LEASE_VALID:
                    cached_leases[self.__cached_lease_key].pop(i)
                    self.__consume_token_lease(lease)
                    return lease
                elif token_status == TokenLeaseStatus.TOKEN_LEASE_EXPIRED:
                    logging.debug(
                        "Discarding token lease '%s' as it is now expired",
                        lease.token,
                    )
                    cached_leases[self.__cached_lease_key].pop(i - num_discarded_leases)
                    num_discarded_leases += 1
                elif token_status == TokenLeaseStatus.TOKEN_LEASE_WRONG_PRIORITY:
                    logging.debug(
                        "Not using token lease '%s' as the priority boost is above "
                        "the configured priority boost (%d <= %d)",
                        lease.token,
                        self.__priority_boost,
                        lease.priority_boost,
                    )

    def __set_cached_token_leases(self, leases: Iterable[quota_pb2.TokenLease]):
        """Replaces all cached token leases with a new set of leases."""

        leases_with_expiration: List[tuple[datetime.datetime, quota_pb2.TokenLease]] = (
            []
        )

        # If Hub doesn't return an expiration date for any of the leases, infer
        # it using the duration_msec field associated with the lease.
        for lease in leases:
            expiration = (
                datetime.datetime.now() + timedelta(milliseconds=lease.duration_msec)
                if lease.expires_at.seconds == 0 and lease.expires_at.nanos == 0
                else (
                    datetime.datetime.fromtimestamp(
                        lease.expires_at.seconds, timezone.utc
                    )
                    + timedelta(microseconds=lease.expires_at.nanos / 2000)
                )
            )
            leases_with_expiration.append((expiration, lease))

        with cached_leases_lock:
            cached_leases[self.__cached_lease_key].extend(leases_with_expiration)

    def __check_token_lease(
        self,
        lease: quota_pb2.TokenLease,
        expiration: datetime.datetime,
    ) -> TokenLeaseStatus:
        """Returns false if a token lease is expired or invalid.

        Both the lease and its expiration are accepted separately. This is done
        since Hub may not specify an expiration date in the lease it returns,
        and we have to create it ourselves. The timestamp fields in the lease
        messages are immutable and cannot be changed at runtime.
        """

        if datetime.datetime.now() >= expiration:
            logging.debug(
                "%s >= %s",
                datetime.datetime.now(),
                expiration,
            )
            return TokenLeaseStatus.TOKEN_LEASE_EXPIRED
        elif self.__priority_boost > lease.priority_boost:
            return TokenLeaseStatus.TOKEN_LEASE_WRONG_PRIORITY
        else:
            return TokenLeaseStatus.TOKEN_LEASE_VALID

    def __consume_token_lease(self, lease: quota_pb2.TokenLease):
        """Marks a token lease as consumed so we can notify Hub later."""

        with consumed_leases_lock:
            consumed_leases[self.__consumed_lease_key].append(lease)

        logging.debug("Currently sitting at %d consumed leases", len(consumed_leases))


def start_batch_token_consumer(event_loop: asyncio.AbstractEventLoop):
    """Continuously run the batch token consumer on the passed in event loop.

    This function should only be wrapped in call_soon_threadsafe() and called
    with the aforementioned event loop.
    """

    def wrapper():
        task = asyncio.create_task(batch_token_consumer())
        task.add_done_callback(handle_batch_token_consumer(event_loop))

    return wrapper


async def batch_token_consumer():
    """Iterates through all consumed tokens and consumes them in batches.

    This function will loop indefinitely by an interval time as defined in
    'BATCH_TOKEN_CONSUME_INTERVAL' (e.g. 200ms).
    """

    global consumed_leases

    logging.debug("Starting batch token consumer background worker")

    while True:
        start_time = time.perf_counter()

        # Quickly snatch and empty out the consumed leases so that we don't block
        # future leases from being consumed while we wait for Hub to finish
        # processing our batch request.
        with consumed_leases_lock:
            seized_leases = consumed_leases
            consumed_leases = defaultdict(list)

        consumption_tasks: List[asyncio.Task] = []

        for environment, leases in seized_leases.items():
            consumption_tasks.append(
                asyncio.create_task(set_token_lease_consumed(leases, environment))
            )

        if len(consumption_tasks) > 0:
            await asyncio.wait(consumption_tasks)

        # We measure the time it takes to consume tokens and subtract that from
        # the interval to ensure that we don't surpass the 200ms deadline as
        # defined in the specification under nominal networking conditions.
        elapsed_time = time.perf_counter() - start_time
        delay = max(BATCH_TOKEN_CONSUME_INTERVAL - elapsed_time, 0)

        if delay > 0:
            await asyncio.sleep(delay)


def handle_batch_token_consumer(event_loop: asyncio.AbstractEventLoop):
    """Gracefully handle and log errors from the background consumer task.

    If the task fails because of an error, it will be rescheduled and continue
    to attempt to send consumed tokens to Hub on an interval.
    """

    # TODO: Retry logic for consumed tokens that we failed to send to Hub?

    def wrapper(task: asyncio.Task):
        global batch_token_consumer_handle

        try:
            task.result()
        except asyncio.CancelledError:
            pass  # Do not log task cancellations.
        except Exception:
            logging.exception("Received unexpected exception while consuming leases")

            with batch_token_consumer_handle_lock:
                batch_token_consumer_handle = event_loop.call_soon_threadsafe(
                    start_batch_token_consumer(event_loop)
                )

    return wrapper


async def set_token_lease_consumed(
    leases: Iterable[quota_pb2.TokenLease],
    environment: str,
):
    """Consume a set of token leases for a given environment."""

    import getstanza.client

    client = getstanza.client.StanzaClient.getInstance()
    if client.hub is None:
        raise RuntimeError("The Stanza SDK has not yet been initialized")

    set_token_lease_consumed_request = quota_pb2.SetTokenLeaseConsumedRequest(
        tokens=list(map(lambda lease: lease.token, leases)),
        environment=environment,
    )
    logging.debug(
        "Consuming %d token leases from environment '%s'",
        len(set_token_lease_consumed_request.tokens),
        environment,
    )

    client.hub.quota_service.SetTokenLeaseConsumed(
        request=set_token_lease_consumed_request,
        metadata=client.config.metadata,
        timeout=DEFAULT_TIMEOUT,
    )



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: stanza/hub/v1/auth.proto
# Protobuf Python Version: 4.25.2
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.api import annotations_pb2 as google_dot_api_dot_annotations__pb2
from protoc_gen_openapiv2.options import annotations_pb2 as protoc__gen__openapiv2_dot_options_dot_annotations__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x18stanza/hub/v1/auth.proto\x12\rstanza.hub.v1\x1a\x1cgoogle/api/annotations.proto\x1a.protoc-gen-openapiv2/options/annotations.proto\"9\n\x15GetBearerTokenRequest\x12 \n\x0b\x65nvironment\x18\x01 \x01(\tR\x0b\x65nvironment\";\n\x16GetBearerTokenResponse\x12!\n\x0c\x62\x65\x61rer_token\x18\x01 \x01(\tR\x0b\x62\x65\x61rerToken2\xaf\x02\n\x0b\x41uthService\x12\x9f\x02\n\x0eGetBearerToken\x12$.stanza.hub.v1.GetBearerTokenRequest\x1a%.stanza.hub.v1.GetBearerTokenResponse\"\xbf\x01\x92\x41\xa5\x01\x12\x10Get Bearer Token\x1aYInspects the X-Stanza-Key auth header and returns a new Bearer Token if API key is valid.J6\n\x03\x32\x30\x30\x12/\n\x02OK\x12)\n\'\x1a%.stanza.hub.v1.GetBearerTokenResponse\x82\xd3\xe4\x93\x02\x10\x12\x0e/v1/auth/tokenB\x89\x01\n\x11\x63om.stanza.hub.v1B\tAuthProtoP\x01Z\x13stanza/hub/v1;hubpb\xa2\x02\x03SHX\xaa\x02\rStanza.Hub.V1\xca\x02\rStanza\\Hub\\V1\xe2\x02\x19Stanza\\Hub\\V1\\GPBMetadata\xea\x02\x0fStanza::Hub::V1b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'stanza.hub.v1.auth_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\021com.stanza.hub.v1B\tAuthProtoP\001Z\023stanza/hub/v1;hubpb\242\002\003SHX\252\002\rStanza.Hub.V1\312\002\rStanza\\Hub\\V1\342\002\031Stanza\\Hub\\V1\\GPBMetadata\352\002\017Stanza::Hub::V1'
  _globals['_AUTHSERVICE'].methods_by_name['GetBearerToken']._options = None
  _globals['_AUTHSERVICE'].methods_by_name['GetBearerToken']._serialized_options = b'\222A\245\001\022\020Get Bearer Token\032YInspects the X-Stanza-Key auth header and returns a new Bearer Token if API key is valid.J6\n\003200\022/\n\002OK\022)\n\'\032%.stanza.hub.v1.GetBearerTokenResponse\202\323\344\223\002\020\022\016/v1/auth/token'
  _globals['_GETBEARERTOKENREQUEST']._serialized_start=121
  _globals['_GETBEARERTOKENREQUEST']._serialized_end=178
  _globals['_GETBEARERTOKENRESPONSE']._serialized_start=180
  _globals['_GETBEARERTOKENRESPONSE']._serialized_end=239
  _globals['_AUTHSERVICE']._serialized_start=242
  _globals['_AUTHSERVICE']._serialized_end=545
# @@protoc_insertion_point(module_scope)



import pytest
from fastapi import HTTPException, Request, status
from getstanza.client import StanzaClient
from getstanza.guard import GuardedStatus
from getstanza_fastapi.fastapi_guard import StanzaGuard


@pytest.fixture(autouse=True)
def context_from_http_headers(monkeypatch):
    def mock_context_from_http_headers():
        return None

    monkeypatch.setattr(StanzaClient, "getInstance", mock_context_from_http_headers)


class FakeGuard:
    def __init__(self, error=None, block_message=None, block_reason=None):
        self.success = GuardedStatus.GUARDED_SUCCESS
        self.failure = GuardedStatus.GUARDED_FAILURE
        self.error = error
        self.block_message = block_message
        self.block_reason = block_reason

    def blocked(self) -> bool:
        return self.block_message is not None

    def end(*args):
        pass


class FakeClient:
    def __init__(self, guard):
        self.__guard = guard

    async def guard(
        self,
        guard_name,
        feature=None,
        priority_boost=None,
        default_weight=None,
        tags=None,
    ):
        return self.__guard


@pytest.fixture()
def stanza_client_patcher(monkeypatch):
    def patcher(fake_guard: FakeGuard) -> None:
        def mock_getInstance():
            return FakeClient(fake_guard)

        monkeypatch.setattr(StanzaClient, "getInstance", mock_getInstance)

    return patcher


@pytest.fixture
def mock_request():
    return Request(scope={"type": "http", "headers": []})


def test_sync_guard_success(stanza_client_patcher, mock_request):
    stanza_client_patcher(FakeGuard())
    try:
        with StanzaGuard(mock_request, "MockGuard"):
            assert True
    except Exception:
        assert False


def test_sync_guard_blocked(mock_request, stanza_client_patcher):
    stanza_client_patcher(
        FakeGuard(block_message="fake block", block_reason="fake reason")
    )
    try:
        with StanzaGuard(mock_request, "MockGuard"):
            assert False
    except HTTPException as req_exc:
        assert req_exc.status_code == status.HTTP_429_TOO_MANY_REQUESTS
        assert req_exc.detail == {
            "message": "fake block",
            "reason": "fake reason",
        }


def test_sync_guard_error(caplog, mock_request, stanza_client_patcher):
    stanza_client_patcher(FakeGuard(error="fake error"))
    try:
        with StanzaGuard(mock_request, "MockGuard"):
            assert True
    except Exception:
        assert False
    finally:
        assert "fake error" in caplog.text


@pytest.mark.asyncio
async def test_async_guard(mock_request, stanza_client_patcher):
    stanza_client_patcher(FakeGuard())
    try:
        async with StanzaGuard(mock_request, "MockGuard"):
            assert True
    except Exception:
        assert False



from collections import defaultdict
from typing import Optional
from unittest.mock import patch

import pytest
from getstanza import guard
from getstanza.guard import Guard
from getstanza.hub import StanzaHub
from getstanza.propagation import context_from_http_headers
from getstanza.tests.utils import async_noop, async_return, noop
from pytest_socket import enable_socket, socket_allow_hosts
from stanza.hub.v1 import common_pb2, config_pb2


@pytest.fixture(autouse=True)
def setup_and_teardown():
    """Clear global cached and consumed lease state between each test."""

    # Block internet access by default for all tests as we mock all calls going
    # out to Hub. Tests will throw 'SocketBlockedError' errors if a mock is
    # missing whilst executing them.
    #
    # We have two notable exceptions. We explicitly allow localhost so that the
    # VSCode debugger doesn't get blocked, and we also allow UNIX sockets since
    # asyncio appears to rely on them.
    socket_allow_hosts(allowed=["localhost"], allow_unix_socket=True)

    # Initialize an empty context for the tests. If we don't do this then we'll
    # get LookupError errors as these tests don't check for incoming baggage.
    context_from_http_headers({})

    yield  # Run the test before executing the teardown logic.

    with guard.cached_leases_lock:
        guard.cached_leases = defaultdict(list)

    with guard.consumed_leases_lock:
        guard.consumed_leases = defaultdict(list)

    enable_socket()


def make_guard(
    hub: StanzaHub,
    guard_name: str,
    feature_name: Optional[str] = None,
    priority_boost: Optional[int] = None,
    default_weight: Optional[float] = None,
    tags=None,
):
    """Create a guard without starting any background tasks."""

    with (
        patch(
            "getstanza.guard.batch_token_consumer_handle"
        ) as batch_token_consumer_handle,
        patch("getstanza.guard.batch_token_consumer") as batch_token_consumer,
        patch(
            "getstanza.guard.handle_batch_token_consumer"
        ) as handle_batch_token_consumer,
    ):
        batch_token_consumer_handle.return_value = async_noop
        batch_token_consumer.return_value = async_noop
        handle_batch_token_consumer.return_value = noop

        guard = Guard(
            hub,
            guard_name,
            feature_name=feature_name,
            priority_boost=priority_boost,
            default_weight=default_weight,
            tags=tags,
        )

        return guard


@pytest.fixture
def quota_guard_config():
    return config_pb2.GuardConfig(
        validate_ingress_tokens=False,
        check_quota=True,
        quota_tags=[],
        report_only=False,
    )


@pytest.fixture
def quota_guard(stanza_hub, quota_guard_config):
    stanza_hub.config_manager.get_guard_config.return_value = async_return(
        (quota_guard_config, common_pb2.Config.CONFIG_CACHED_OK)
    )
    return make_guard(
        stanza_hub,
        "QuotaGuard",
        feature_name=None,
        priority_boost=0,
        tags=None,
    )


@pytest.fixture
def quota_guard_with_feature(stanza_hub, quota_guard_config):
    stanza_hub.config_manager.get_guard_config.return_value = async_return(
        (quota_guard_config, common_pb2.Config.CONFIG_CACHED_OK)
    )
    return make_guard(
        stanza_hub,
        "QuotaGuard",
        feature_name="QuotaGuardFeature",
        priority_boost=0,
        tags=None,
    )


@pytest.fixture
def quota_guard_with_priority_boost(stanza_hub, quota_guard_config):
    stanza_hub.config_manager.get_guard_config.return_value = async_return(
        (quota_guard_config, common_pb2.Config.CONFIG_CACHED_OK)
    )
    return make_guard(
        stanza_hub,
        "QuotaGuard",
        feature_name=None,
        priority_boost=5,
        tags=None,
    )


@pytest.fixture
def guard_without_config(stanza_hub):
    stanza_hub.config_manager.get_guard_config.return_value = async_return(
        (None, common_pb2.Config.CONFIG_UNSPECIFIED)
    )
    return make_guard(
        stanza_hub,
        "GuardWithoutConfig",
        feature_name=None,
        priority_boost=0,
        tags=None,
    )


@pytest.fixture
def guard_without_config_fetch_error(stanza_hub):
    stanza_hub.config_manager.get_guard_config.return_value = async_return(
        (None, common_pb2.Config.CONFIG_FETCH_ERROR)
    )
    return make_guard(
        stanza_hub,
        "GuardWithoutConfig",
        feature_name=None,
        priority_boost=0,
        tags=None,
    )


@pytest.fixture
def guard_without_config_fetch_timeout(stanza_hub):
    stanza_hub.config_manager.get_guard_config.return_value = async_return(
        (None, common_pb2.Config.CONFIG_FETCH_TIMEOUT)
    )
    return make_guard(
        stanza_hub,
        "GuardWithoutConfig",
        feature_name=None,
        priority_boost=0,
        tags=None,
    )


@pytest.fixture
def guard_without_config_not_found(stanza_hub):
    stanza_hub.config_manager.get_guard_config.return_value = async_return(
        (None, common_pb2.Config.CONFIG_NOT_FOUND)
    )
    return make_guard(
        stanza_hub,
        "GuardWithoutConfig",
        feature_name=None,
        priority_boost=0,
        tags=None,
    )


@pytest.fixture
def token_guard_config():
    return config_pb2.GuardConfig(
        validate_ingress_tokens=True,
        check_quota=False,
        quota_tags=[],
        report_only=False,
    )


@pytest.fixture
def token_guard(stanza_hub, token_guard_config):
    stanza_hub.config_manager.get_guard_config.return_value = async_return(
        (token_guard_config, common_pb2.Config.CONFIG_CACHED_OK)
    )
    return make_guard(
        stanza_hub,
        "TokenGuard",
        feature_name=None,
        priority_boost=0,
        tags=None,
    )


@pytest.fixture
def report_only_guard_config():
    return config_pb2.GuardConfig(
        validate_ingress_tokens=False,
        check_quota=True,
        quota_tags=[],
        report_only=True,
    )


@pytest.fixture
def report_only_guard(stanza_hub, report_only_guard_config):
    stanza_hub.config_manager.get_guard_config.return_value = async_return(
        (report_only_guard_config, common_pb2.Config.CONFIG_CACHED_OK)
    )
    return make_guard(
        stanza_hub,
        "ReportOnlyQuotaGuard",
        feature_name=None,
        priority_boost=0,
        tags=None,
    )


# TODO: Ingress token report only as well?



import asyncio
import logging
import random
from typing import Optional

from getstanza.hub.configuration_manager import StanzaHubConfigurationManager

# Set to less than the maximum duration of the Auth0 Bearer Token
BEARER_TOKEN_REFRESH_INTERVAL_SECS = 43_000
BEARER_TOKEN_REFRESH_JITTER_SECS = 200


class StanzaTelemetryCollector:
    """Implements reconnecting to Stanza's OpenTelemetry collector."""

    def __init__(
        self,
        event_loop: asyncio.AbstractEventLoop,
        config_manager: StanzaHubConfigurationManager,
    ):
        self.event_loop = event_loop
        self.config_manager = config_manager

        self.__task: Optional[asyncio.Task[None]] = None
        self.event_loop.call_later(self.__jitter(), self.__schedule)

    def __jitter(self) -> float:
        return BEARER_TOKEN_REFRESH_INTERVAL_SECS + random.randint(
            0, BEARER_TOKEN_REFRESH_JITTER_SECS
        )

    async def __connect(self):
        await asyncio.wait([asyncio.create_task(self.config_manager.connect_otel())])

    async def __interval(self):
        try:
            await self.__connect()
        finally:
            self.event_loop.call_later(self.__jitter(), self.__schedule)

    def __schedule(self):
        self.__task = asyncio.create_task(self.__interval())
        self.__task.add_done_callback(self.__handle_result)

    def __handle_result(self, task: asyncio.Task):
        try:
            task.result()
        except asyncio.CancelledError:
            pass  # Do not log task cancellations.
        except Exception as exc:
            logging.exception(
                "Received unexpected exception while connecting to OTEL collector: %s",
                exc,
            )



import logging
import os

from getstanza.otel.meter import StanzaMeter, StanzaMeterProvider
from getstanza.otel.tracer import StanzaTracerProvider
from opentelemetry.sdk.resources import Resource
from opentelemetry.sdk.trace.sampling import TraceIdRatioBased
from opentelemetry.trace import Tracer

INSTRUMENTATION_NAME = "github.com/StanzaSystems/sdk-python"
INSTRUMENTATION_VERSION = "0.0.1-beta"


class OpenTelemetry:
    """Implements OTEL metrics and traces."""

    @property
    def meter(self) -> StanzaMeter:
        return self.__meter

    @property
    def tracer(self) -> Tracer:
        return self.__tracer

    def __init__(
        self,
        bearer_token: str,
        metric_collector_url: str,
        trace_collector_url: str,
        trace_sample_rate: float,
        service_name: str,
        service_release: str,
        environment: str,
    ):
        self.__debug = False
        self.__insecure = False
        self.__metric_collector_url = metric_collector_url
        self.__trace_collector_url = trace_collector_url
        self.__trace_sample_rate = trace_sample_rate

        if os.environ.get("STANZA_OTEL_DEBUG"):
            self.__debug = True

        if os.environ.get("STANZA_OTEL_INSECURE"):
            self.__insecure = True

        # TODO: Fetch bearer token here instead of passing it in?
        self.__bearer_token = bearer_token

        # Define shared OpenTelemetry resource.
        # https://opentelemetry.io/docs/specs/otel/resource/sdk/
        self.__resource = Resource.create(
            {
                "service.name": service_name,
                "service.version": service_release,
                "deployment.environment": environment,
            }
        )

    def new_meter(self) -> bool:
        """Setup OpenTelemetry meter provider and create new meter."""
        try:
            self.__meter_provider = StanzaMeterProvider(
                self.__debug,
                self.__insecure,
                resource=self.__resource,
                endpoint=self.__metric_collector_url,
                headers={"authorization": "Bearer " + self.__bearer_token},
            )
            self.__meter = self.__meter_provider.get_meter(
                INSTRUMENTATION_NAME, INSTRUMENTATION_VERSION
            )
            return True
        except Exception as exc:
            # TODO: add real/better exception handling
            logging.exception(exc)
            return False

    def new_tracer(self) -> bool:
        """Setup OpenTelemetry trace provider."""
        try:
            self.__trace_provider = StanzaTracerProvider(
                debug=self.__debug,
                insecure=self.__insecure,
                resource=self.__resource,
                endpoint=self.__trace_collector_url,
                sampler=TraceIdRatioBased(self.__trace_sample_rate),
                headers={"authorization": "Bearer " + self.__bearer_token},
            )
            self.__tracer = self.__trace_provider.get_tracer(
                INSTRUMENTATION_NAME, INSTRUMENTATION_VERSION
            )
            return True
        except Exception as exc:
            # TODO: add real/better exception handling
            logging.exception(exc)
            return False



import asyncio
import logging
import re
import traceback
from typing import Dict, Optional

from getstanza.client import StanzaClient
from getstanza.guard import Guard
from getstanza.propagation import context_from_mapping
from getstanza_sqs.exceptions import GuardBlockedError

QUEUE_ARN_RE = r"^arn:aws:sqs:([a-zA-Z0-9-]+):([0-9]+):(.+)$"


class SQSGuard:
    """
    Wrapper around boto3 resource objects for intercepting calls.

    We don't extend directly from the Queue resource class as it is generated
    at runtime, and we can't tell boto3 to use a different class. To workaround
    this we wrap the queue resource instead and passthrough method calls to it.
    """

    def __init__(
        self,
        queue,
        guard_name: str,
        feature_name: Optional[str] = None,
        priority_boost: Optional[int] = None,
        default_weight: Optional[float] = None,
        tags: Optional[Dict[str, str]] = None,
    ):
        self.__queue = queue
        self.__guard_name = guard_name
        self.__feature_name = feature_name
        self.__priority_boost = priority_boost
        self.__default_weight = default_weight
        self.__tags = tags
        self.__client = StanzaClient.getInstance()

    def __getattribute__(self, name):
        try:
            return super().__getattribute__(name)
        except AttributeError:
            return getattr(self.__queue, name)

    def __process_message(self, message):
        """
        Guard incoming SQS messages.

        In the event that a guard blocks, we terminate the visibility timeout
        so that the message can be processed again later.
        """

        if not self.__client.hub:
            return message  # Passthrough if SDK isn't initialized yet.

        context_from_mapping(
            dict(
                (key, value["StringValue"])
                for key, value in (message.message_attributes or {}).items()
                if value["DataType"] == "String"
            )
        )

        guard = Guard(
            self.__client.hub,
            self.__guard_name,
            feature_name=self.__feature_name,
            priority_boost=self.__priority_boost,
            default_weight=self.__default_weight,
            tags=self.__tags,
        )
        asyncio.run_coroutine_threadsafe(
            guard.run(), self.__client.hub.event_loop
        ).result()

        # 🪵 Check for and log any returned error messages
        if guard.error:
            logging.error(guard.error)

        # 🚫 Stanza Guard has *blocked* this workflow log the error and
        # raise an HTTPException with a 429 response code.
        if guard.blocked():
            logging.error(guard.block_message, extra={"reason": guard.block_reason})
            raise GuardBlockedError(guard_name=self.__guard_name)

        return message

    def __reset_messages(self, messages):
        """
        Terminate the visibility timeout on a message. This should be done for
        messages that we can't process due to an error in processing the
        message, or because the guard blocked it.
        """

        logging.debug("Resetting visibility timeout for %d message(s)", len(messages))

        response = self.__queue.meta.client.change_message_visibility_batch(
            QueueUrl=self.__queue_arn_to_url(self.__queue.attributes["QueueArn"]),
            Entries=[
                {
                    "Id": message.message_id,
                    "ReceiptHandle": message.receipt_handle,
                    "VisibilityTimeout": 1,  # TODO: Make this configurable
                }
                for message in messages
            ],
        )

        for failed_message in response.get("Failed", ()):
            logging.error(
                "Unable to reset visibility timeout for SQS message "
                "(Id=%s, SenderFault=%s, Code=%s, Message=%s)",
                failed_message["Id"],
                failed_message["SenderFault"],
                failed_message["Code"],
                failed_message["Message"],
            )

        return response

    def __queue_arn_to_url(self, arn: str) -> str:
        """
        Converts an SQS queue ARN to an HTTP URL.
        """

        if result := re.match(QUEUE_ARN_RE, arn):
            region, account_id, resource_id = result.groups()
            return f"https://sqs.{region}.amazonaws.com/{account_id}/{resource_id}"

        raise ValueError(f"{arn} is not a valid ARN")

    def receive_messages(self, *args, **kwargs):
        """
        Retrieves one or more messages (up to 10), from the specified queue.
        Using the WaitTimeSeconds parameter enables long-poll support.
        """

        messages = self.__queue.receive_messages(*args, **kwargs)

        if not self.__client.hub:
            return messages  # Passthrough if SDK isn't initialized yet.

        successful_messages = []
        failed_messages = []

        for message in messages:
            try:
                successful_messages.append(self.__process_message(message))
            except GuardBlockedError:
                failed_messages.append(message)
            except Exception:
                logging.error(
                    "Unable to process SQS message: %s", traceback.format_exc()
                )
                failed_messages.append(message)

        if len(failed_messages) > 0:
            self.__reset_messages(failed_messages)

        return successful_messages



import pytest
from getstanza.configuration import StanzaConfiguration

TEST_API_KEY = "0000000000000000000000000000000000000000000000000000000000000000"
TEST_SERVICE_NAME = "TestService"
TEST_SERVICE_RELEASE = "0.0.0"
TEST_ENVIRONMENT = "dev"
TEST_HUB_ADDRESS = "127.0.0.1:9060"


@pytest.fixture
def stanza_config():
    return StanzaConfiguration(
        api_key=TEST_API_KEY,
        service_name=TEST_SERVICE_NAME,
        service_release=TEST_SERVICE_RELEASE,
        environment=TEST_ENVIRONMENT,
        hub_address=TEST_HUB_ADDRESS,
    )



"""
An adapter to ease integration of Stanza with FastAPI applications.
"""

from .fastapi_client import StanzaFastAPIClient
from .fastapi_guard import StanzaGuard

__all__ = ["StanzaFastAPIClient", "StanzaGuard"]



import asyncio
import os

import grpc
from getstanza.configuration import StanzaConfiguration
from getstanza.hub.configuration_manager import StanzaHubConfigurationManager
from getstanza.hub.poller import StanzaHubPoller
from getstanza.hub.telemetry import StanzaTelemetryCollector
from stanza.hub.v1 import auth_pb2_grpc, config_pb2_grpc, quota_pb2_grpc


class StanzaHub:
    """Implements all SDK to Stanza Hub interactions."""

    def __init__(self, config: StanzaConfiguration) -> None:
        # TODO: Send user-agent per SDK spec
        if os.environ.get("STANZA_HUB_NO_TLS"):  # Disable TLS for local Hub development
            self.__grpc_channel = grpc.insecure_channel(config.hub_address)
        else:
            self.__grpc_channel = grpc.secure_channel(
                config.hub_address, grpc.ssl_channel_credentials()
            )

        try:
            asyncio.get_running_loop()
        except RuntimeError:
            pass  # Expected result as we want no event loop running yet.
        else:
            raise RuntimeError(
                "Cannot create an event loop for Hub as this thread already "
                "has a loop initialized."
            )

        # Other threads will need to interact with Stanza. We keep a reference
        # to the loop being used by the SDK client from the thread it was
        # initialized on to ensure that any new recurring tasks spawned from
        # other threads are able to run for the duration of the SDK client's
        # lifetime.
        self.event_loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.event_loop)

        # Create Stanza Hub service stubs.
        self.auth_service = auth_pb2_grpc.AuthServiceStub(self.__grpc_channel)
        self.config_service = config_pb2_grpc.ConfigServiceStub(self.__grpc_channel)
        self.quota_service = quota_pb2_grpc.QuotaServiceStub(self.__grpc_channel)

        # Create initial configuration and poller.
        self.config_manager = StanzaHubConfigurationManager(
            auth_service=self.auth_service,
            config_service=self.config_service,
            config=config,
        )
        self.hub_poller = StanzaHubPoller(
            event_loop=self.event_loop,
            config_manager=self.config_manager,
            interval=config.interval,
        )
        self.telemetry_collector = StanzaTelemetryCollector(
            event_loop=self.event_loop,
            config_manager=self.config_manager,
        )

    def start_poller(self):
        """Start async polling of Hub for Service and Guard configs"""

        self.hub_poller.start()
        self.event_loop.run_forever()



from google.api import annotations_pb2 as _annotations_pb2
from protoc_gen_openapiv2.options import annotations_pb2 as _annotations_pb2_1
from stanza.hub.v1 import common_pb2 as _common_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class GetGuardConfigRequest(_message.Message):
    __slots__ = ("version_seen", "selector")
    VERSION_SEEN_FIELD_NUMBER: _ClassVar[int]
    SELECTOR_FIELD_NUMBER: _ClassVar[int]
    version_seen: str
    selector: _common_pb2.GuardServiceSelector
    def __init__(self, version_seen: _Optional[str] = ..., selector: _Optional[_Union[_common_pb2.GuardServiceSelector, _Mapping]] = ...) -> None: ...

class GetGuardConfigResponse(_message.Message):
    __slots__ = ("version", "config_data_sent", "config")
    VERSION_FIELD_NUMBER: _ClassVar[int]
    CONFIG_DATA_SENT_FIELD_NUMBER: _ClassVar[int]
    CONFIG_FIELD_NUMBER: _ClassVar[int]
    version: str
    config_data_sent: bool
    config: GuardConfig
    def __init__(self, version: _Optional[str] = ..., config_data_sent: bool = ..., config: _Optional[_Union[GuardConfig, _Mapping]] = ...) -> None: ...

class GuardConfig(_message.Message):
    __slots__ = ("validate_ingress_tokens", "check_quota", "quota_tags", "report_only")
    VALIDATE_INGRESS_TOKENS_FIELD_NUMBER: _ClassVar[int]
    CHECK_QUOTA_FIELD_NUMBER: _ClassVar[int]
    QUOTA_TAGS_FIELD_NUMBER: _ClassVar[int]
    REPORT_ONLY_FIELD_NUMBER: _ClassVar[int]
    validate_ingress_tokens: bool
    check_quota: bool
    quota_tags: _containers.RepeatedScalarFieldContainer[str]
    report_only: bool
    def __init__(self, validate_ingress_tokens: bool = ..., check_quota: bool = ..., quota_tags: _Optional[_Iterable[str]] = ..., report_only: bool = ...) -> None: ...

class GetBrowserContextRequest(_message.Message):
    __slots__ = ("feature",)
    FEATURE_FIELD_NUMBER: _ClassVar[int]
    feature: _common_pb2.FeatureSelector
    def __init__(self, feature: _Optional[_Union[_common_pb2.FeatureSelector, _Mapping]] = ...) -> None: ...

class GetBrowserContextResponse(_message.Message):
    __slots__ = ("feature_configs",)
    FEATURE_CONFIGS_FIELD_NUMBER: _ClassVar[int]
    feature_configs: _containers.RepeatedCompositeFieldContainer[FeatureConfig]
    def __init__(self, feature_configs: _Optional[_Iterable[_Union[FeatureConfig, _Mapping]]] = ...) -> None: ...

class FeatureConfig(_message.Message):
    __slots__ = ("name", "config")
    NAME_FIELD_NUMBER: _ClassVar[int]
    CONFIG_FIELD_NUMBER: _ClassVar[int]
    name: str
    config: BrowserConfig
    def __init__(self, name: _Optional[str] = ..., config: _Optional[_Union[BrowserConfig, _Mapping]] = ...) -> None: ...

class BrowserConfig(_message.Message):
    __slots__ = ("enabled_percent", "action_code_enabled", "message_enabled", "action_code_disabled", "message_disabled")
    ENABLED_PERCENT_FIELD_NUMBER: _ClassVar[int]
    ACTION_CODE_ENABLED_FIELD_NUMBER: _ClassVar[int]
    MESSAGE_ENABLED_FIELD_NUMBER: _ClassVar[int]
    ACTION_CODE_DISABLED_FIELD_NUMBER: _ClassVar[int]
    MESSAGE_DISABLED_FIELD_NUMBER: _ClassVar[int]
    enabled_percent: int
    action_code_enabled: int
    message_enabled: str
    action_code_disabled: int
    message_disabled: str
    def __init__(self, enabled_percent: _Optional[int] = ..., action_code_enabled: _Optional[int] = ..., message_enabled: _Optional[str] = ..., action_code_disabled: _Optional[int] = ..., message_disabled: _Optional[str] = ...) -> None: ...

class GetServiceConfigRequest(_message.Message):
    __slots__ = ("version_seen", "service", "client_id")
    VERSION_SEEN_FIELD_NUMBER: _ClassVar[int]
    SERVICE_FIELD_NUMBER: _ClassVar[int]
    CLIENT_ID_FIELD_NUMBER: _ClassVar[int]
    version_seen: str
    service: _common_pb2.ServiceSelector
    client_id: str
    def __init__(self, version_seen: _Optional[str] = ..., service: _Optional[_Union[_common_pb2.ServiceSelector, _Mapping]] = ..., client_id: _Optional[str] = ...) -> None: ...

class GetServiceConfigResponse(_message.Message):
    __slots__ = ("version", "config_data_sent", "config")
    VERSION_FIELD_NUMBER: _ClassVar[int]
    CONFIG_DATA_SENT_FIELD_NUMBER: _ClassVar[int]
    CONFIG_FIELD_NUMBER: _ClassVar[int]
    version: str
    config_data_sent: bool
    config: ServiceConfig
    def __init__(self, version: _Optional[str] = ..., config_data_sent: bool = ..., config: _Optional[_Union[ServiceConfig, _Mapping]] = ...) -> None: ...

class ServiceConfig(_message.Message):
    __slots__ = ("customer_id", "trace_config", "metric_config", "sentinel_config")
    CUSTOMER_ID_FIELD_NUMBER: _ClassVar[int]
    TRACE_CONFIG_FIELD_NUMBER: _ClassVar[int]
    METRIC_CONFIG_FIELD_NUMBER: _ClassVar[int]
    SENTINEL_CONFIG_FIELD_NUMBER: _ClassVar[int]
    customer_id: str
    trace_config: TraceConfig
    metric_config: MetricConfig
    sentinel_config: SentinelConfig
    def __init__(self, customer_id: _Optional[str] = ..., trace_config: _Optional[_Union[TraceConfig, _Mapping]] = ..., metric_config: _Optional[_Union[MetricConfig, _Mapping]] = ..., sentinel_config: _Optional[_Union[SentinelConfig, _Mapping]] = ...) -> None: ...

class TraceConfig(_message.Message):
    __slots__ = ("collector_url", "sample_rate_default", "overrides", "header_sample_configs", "param_sample_configs")
    COLLECTOR_URL_FIELD_NUMBER: _ClassVar[int]
    SAMPLE_RATE_DEFAULT_FIELD_NUMBER: _ClassVar[int]
    OVERRIDES_FIELD_NUMBER: _ClassVar[int]
    HEADER_SAMPLE_CONFIGS_FIELD_NUMBER: _ClassVar[int]
    PARAM_SAMPLE_CONFIGS_FIELD_NUMBER: _ClassVar[int]
    collector_url: str
    sample_rate_default: float
    overrides: _containers.RepeatedCompositeFieldContainer[TraceConfigOverride]
    header_sample_configs: _containers.RepeatedCompositeFieldContainer[HeaderTraceConfig]
    param_sample_configs: _containers.RepeatedCompositeFieldContainer[ParamTraceConfig]
    def __init__(self, collector_url: _Optional[str] = ..., sample_rate_default: _Optional[float] = ..., overrides: _Optional[_Iterable[_Union[TraceConfigOverride, _Mapping]]] = ..., header_sample_configs: _Optional[_Iterable[_Union[HeaderTraceConfig, _Mapping]]] = ..., param_sample_configs: _Optional[_Iterable[_Union[ParamTraceConfig, _Mapping]]] = ...) -> None: ...

class MetricConfig(_message.Message):
    __slots__ = ("collector_url",)
    COLLECTOR_URL_FIELD_NUMBER: _ClassVar[int]
    collector_url: str
    def __init__(self, collector_url: _Optional[str] = ...) -> None: ...

class SentinelConfig(_message.Message):
    __slots__ = ("circuitbreaker_rules_json", "flow_rules_json", "isolation_rules_json", "system_rules_json")
    CIRCUITBREAKER_RULES_JSON_FIELD_NUMBER: _ClassVar[int]
    FLOW_RULES_JSON_FIELD_NUMBER: _ClassVar[int]
    ISOLATION_RULES_JSON_FIELD_NUMBER: _ClassVar[int]
    SYSTEM_RULES_JSON_FIELD_NUMBER: _ClassVar[int]
    circuitbreaker_rules_json: str
    flow_rules_json: str
    isolation_rules_json: str
    system_rules_json: str
    def __init__(self, circuitbreaker_rules_json: _Optional[str] = ..., flow_rules_json: _Optional[str] = ..., isolation_rules_json: _Optional[str] = ..., system_rules_json: _Optional[str] = ...) -> None: ...

class TraceConfigOverride(_message.Message):
    __slots__ = ("sample_rate", "span_selectors")
    SAMPLE_RATE_FIELD_NUMBER: _ClassVar[int]
    SPAN_SELECTORS_FIELD_NUMBER: _ClassVar[int]
    sample_rate: float
    span_selectors: _containers.RepeatedCompositeFieldContainer[SpanSelector]
    def __init__(self, sample_rate: _Optional[float] = ..., span_selectors: _Optional[_Iterable[_Union[SpanSelector, _Mapping]]] = ...) -> None: ...

class SpanSelector(_message.Message):
    __slots__ = ("otel_attribute", "value")
    OTEL_ATTRIBUTE_FIELD_NUMBER: _ClassVar[int]
    VALUE_FIELD_NUMBER: _ClassVar[int]
    otel_attribute: str
    value: str
    def __init__(self, otel_attribute: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...

class HeaderTraceConfig(_message.Message):
    __slots__ = ("span_selectors", "request_header_names", "response_header_names")
    SPAN_SELECTORS_FIELD_NUMBER: _ClassVar[int]
    REQUEST_HEADER_NAMES_FIELD_NUMBER: _ClassVar[int]
    RESPONSE_HEADER_NAMES_FIELD_NUMBER: _ClassVar[int]
    span_selectors: _containers.RepeatedCompositeFieldContainer[SpanSelector]
    request_header_names: _containers.RepeatedScalarFieldContainer[str]
    response_header_names: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, span_selectors: _Optional[_Iterable[_Union[SpanSelector, _Mapping]]] = ..., request_header_names: _Optional[_Iterable[str]] = ..., response_header_names: _Optional[_Iterable[str]] = ...) -> None: ...

class ParamTraceConfig(_message.Message):
    __slots__ = ("span_selectors", "parameter_names")
    SPAN_SELECTORS_FIELD_NUMBER: _ClassVar[int]
    PARAMETER_NAMES_FIELD_NUMBER: _ClassVar[int]
    span_selectors: _containers.RepeatedCompositeFieldContainer[SpanSelector]
    parameter_names: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, span_selectors: _Optional[_Iterable[_Union[SpanSelector, _Mapping]]] = ..., parameter_names: _Optional[_Iterable[str]] = ...) -> None: ...



from google.api import annotations_pb2 as _annotations_pb2
from protoc_gen_openapiv2.options import annotations_pb2 as _annotations_pb2_1
from stanza.hub.v1 import common_pb2 as _common_pb2
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class QueryGuardHealthRequest(_message.Message):
    __slots__ = ("selector", "priority_boost")
    SELECTOR_FIELD_NUMBER: _ClassVar[int]
    PRIORITY_BOOST_FIELD_NUMBER: _ClassVar[int]
    selector: _common_pb2.GuardFeatureSelector
    priority_boost: int
    def __init__(self, selector: _Optional[_Union[_common_pb2.GuardFeatureSelector, _Mapping]] = ..., priority_boost: _Optional[int] = ...) -> None: ...

class QueryGuardHealthResponse(_message.Message):
    __slots__ = ("health",)
    HEALTH_FIELD_NUMBER: _ClassVar[int]
    health: _common_pb2.Health
    def __init__(self, health: _Optional[_Union[_common_pb2.Health, str]] = ...) -> None: ...



from botocore.exceptions import BotoCoreError


class GuardError(BotoCoreError):
    """Raised when a guard fails with an error."""

    fmt = "An unspecified guard error occurred"


class GuardBlockedError(BotoCoreError):
    """Raised when a blocked guard is preventing a request from going out."""

    fmt = "Stanza blocked guard {guard_name}"



from unittest.mock import patch

import pytest


@pytest.fixture
def stanza_hub(config_manager, quota_service):
    with patch("getstanza.hub.StanzaHub") as MockClass:
        instance = MockClass.return_value
        instance.quota_service = quota_service
        instance.config_manager = config_manager

        yield instance


@pytest.fixture
def config_manager(stanza_config):
    with patch("getstanza.hub.StanzaHubConfigurationManager") as MockClass:
        instance = MockClass.return_value
        instance.config = stanza_config
        instance.get_guard_config.return_value = None

        yield instance


@pytest.fixture
def quota_service():
    with patch("stanza.hub.v1.quota_pb2_grpc.QuotaServiceStub") as MockClass:
        instance = MockClass.return_value
        instance.GetToken.return_value = None
        instance.GetTokenLease.return_value = None
        instance.SetTokenLeaseConsumed.return_value = None
        instance.ValidateToken.return_value = None

        yield instance



import asyncio
import logging
from types import TracebackType
from typing import Dict, Optional

from fastapi import HTTPException, Request, status
from getstanza.client import StanzaClient
from getstanza.guard import GuardedStatus
from getstanza.propagation import context_from_http_headers

# TODO: Check if errors can be returned from FastAPI without exceptions, and if
# so then consider checking for that case and emitting a fail in that scenario.
#
# We may need to handle 429 errors differently. For example, if the
# server is using a custom data envelope format for the error, we want to be
# able to use theirs instead of hardcoding it into the application.


class StanzaGuard:
    """Helps wrap a FastAPI request handler with a guard.

    Implementation is derived from the contextlib.ContextDecorator with
    some additional modifications to support our use-case.
    """

    def __init__(
        self,
        request: Request,
        guard_name: str,
        feature_name: Optional[str] = None,
        priority_boost: Optional[int] = None,
        tags: Optional[Dict[str, str]] = None,
    ):
        self.__request = request
        self.__guard_name = guard_name
        self.__feature_name = feature_name
        self.__priority_boost = priority_boost
        self.__tags = tags
        self.__client = StanzaClient.getInstance()

    def __enter__(self):
        event_loop = None
        try:
            event_loop = asyncio.get_running_loop()
        except RuntimeError:
            pass  # This is fine and means that no loop is running.

        if event_loop is not None:
            raise RuntimeError(
                "A StanzaGuard context manager cannot be used synchronously in "
                "an asynchronous context. Please use as follows:\n\n"
                "'async with StanzaGuard(request, guard_name, ...)'"
            )

        return asyncio.run(self._execute_guard())

    def __exit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ):
        exit_status = self.__guard.success if exc_val is None else self.__guard.failure
        message = str(exc_val) if exc_val else None

        return self._end_guard(exit_status, message)

    async def __aenter__(self):
        return await self._execute_guard()

    async def __aexit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ):
        exit_status = self.__guard.success if exc_val is None else self.__guard.failure
        message = str(exc_val) if exc_val else None

        return self._end_guard(exit_status, message)

    async def _execute_guard(self):
        """
        Runs the guard this context manager was made for. If the guard blocks
        then an HTTPException with HTTP code 429 will be raised.
        """

        logging.debug(
            "Entering guard (name=%s, feature=%s, priority_boost=%s, tags=%r)",
            self.__guard_name,
            self.__feature_name,
            self.__priority_boost,
            self.__tags,
        )

        # Store baggage in a context local to the async handler we're in.
        context_from_http_headers(self.__request.headers)

        self.__guard = await self.__client.guard(
            self.__guard_name,
            feature=self.__feature_name,
            priority_boost=self.__priority_boost,
            tags=self.__tags,
        )

        # 🪵 Check for and log any returned error messages
        if self.__guard.error:
            logging.error(self.__guard.error)

        # 🚫 Stanza Guard has *blocked* this workflow log the error and raise
        # an HTTPException with a 429 response code.
        if self.__guard.blocked():
            logging.error(
                self.__guard.block_message,
                extra={"reason": self.__guard.block_reason},
            )

            raise HTTPException(
                status_code=status.HTTP_429_TOO_MANY_REQUESTS,
                detail={
                    "message": self.__guard.block_message,
                    "reason": self.__guard.block_reason,
                },
            )

        # ✅ Stanza Guard has *allowed* this workflow, business logic follows.
        return self

    def _end_guard(self, exit_status: GuardedStatus, message: Optional[str]):
        """End the guard and collect metrics so that can be communicated."""

        if message:
            logging.debug("Exiting with status '%s': %s", exit_status.name, message)
        else:
            logging.debug("Exiting with status '%s'", exit_status.name)

        self.__guard.end(exit_status)



import datetime
import os
import uuid
from typing import Optional

CONFIG_POLL_INTERVAL_SECS = 15


class StanzaConfiguration:
    """
    Stanza SDK configuration. Values can be set directly as parameters to the
    constructor, and when not present, are pulled from the environment.
    """

    def __init__(
        self,
        api_key: Optional[str] = None,
        service_name: Optional[str] = None,
        service_release: Optional[str] = None,
        environment: Optional[str] = None,
        hub_address: Optional[str] = None,
    ):
        if not api_key:
            if os.environ.get("STANZA_API_KEY"):
                self.api_key = os.environ.get("STANZA_API_KEY")
            else:
                raise ValueError(
                    "Missing required Stanza API key "
                    "(Hint: Set a STANZA_API_KEY environment variable!)"
                )
        else:
            self.api_key = api_key

        if not service_name:
            if os.environ.get("STANZA_SERVICE_NAME"):
                self.service_name = os.environ.get("STANZA_SERVICE_NAME")
            else:
                raise ValueError(
                    "Missing required service name "
                    "(Hint: Set a STANZA_SERVICE_NAME environment variable!)"
                )
        else:
            self.service_name = service_name

        if not service_release:
            if os.environ.get("STANZA_SERVICE_RELEASE"):
                self.service_release = os.environ.get("STANZA_SERVICE_RELEASE")
            else:
                self.service_release = "0.0.0"
        else:
            self.service_release = service_release

        if not environment:
            if os.environ.get("STANZA_ENVIRONMENT"):
                self.environment = os.environ.get("STANZA_ENVIRONMENT")
            else:
                self.environment = "dev"
        else:
            self.environment = environment

        if not hub_address:
            if os.environ.get("STANZA_HUB_ADDRESS"):
                self.hub_address = os.environ.get("STANZA_HUB_ADDRESS")
            else:
                self.hub_address = "hub.stanzasys.co:9020"
        else:
            self.hub_address = hub_address

        self.interval = datetime.timedelta(seconds=CONFIG_POLL_INTERVAL_SECS)
        self.client_id = str(uuid.uuid4())
        self.customer_id: Optional[str] = None
        self.metadata = [("x-stanza-key", self.api_key)]



import asyncio
import datetime
import logging
from typing import Optional

from getstanza.hub.configuration_manager import StanzaHubConfigurationManager


class StanzaHubPoller:
    """Implements polling behavior for re-fetching configuration from Hub."""

    def __init__(
        self,
        event_loop: asyncio.AbstractEventLoop,
        config_manager: StanzaHubConfigurationManager,
        interval: datetime.timedelta,
    ):
        self.event_loop = event_loop
        self.config_manager = config_manager
        self.interval = interval

        self.__polling_task: Optional[asyncio.Task[None]] = None
        self.__polling = False

    def start(self):
        """Begin polling with the configured interval."""

        if not self.__polling:
            self.__polling = True
            self.event_loop.call_soon(self.__schedule_poll)

    def stop(self):
        """Stop polling Hub for configuration changes."""

        if self.__polling_task:
            self.__polling_task.cancel()
            self.__polling = False

    async def __poll(self):
        """Polls for updated configuration information from Hub."""

        configuration_fetch_tasks = [
            asyncio.create_task(self.config_manager.fetch_service_config()),
            asyncio.create_task(self.config_manager.refetch_known_guard_configs()),
        ]
        await asyncio.wait(configuration_fetch_tasks)

    async def __poll_interval(self, _task: Optional[asyncio.Task[None]] = None):
        """Poll Hub then schedule another poll in the future using interval."""

        try:
            if self.__polling:
                await self.__poll()
        finally:
            if self.__polling:
                self.event_loop.call_later(
                    self.interval.total_seconds(), self.__schedule_poll
                )

    def __schedule_poll(self):
        self.__polling_task = asyncio.create_task(self.__poll_interval())
        self.__polling_task.add_done_callback(self.__handle_hub_poll_result)

    def __handle_hub_poll_result(self, task: asyncio.Task):
        try:
            task.result()
        except asyncio.CancelledError:
            pass  # Do not log task cancellations.
        except Exception as exc:
            logging.exception(
                "Received unexpected exception while polling Hub: %s", exc
            )



from unittest.mock import patch

import pytest
from fastapi import FastAPI, Request
from fastapi.testclient import TestClient
from getstanza.configuration import StanzaConfiguration
from getstanza.tests.utils import async_return
from getstanza_fastapi.fastapi_client import StanzaFastAPIClient


@pytest.fixture
def fastapi_guard():
    with patch("getstanza_fastapi.fastapi_guard.StanzaGuard") as MockClass:
        instance = MockClass.return_value
        instance.__aenter__.return_value = async_return(None)
        instance.__aexit__.return_value = async_return(None)
        instance.__enter__.return_value = None
        instance.__exit__.return_value = None

        yield instance


@pytest.fixture(scope="module")
def fastapi_client():
    return StanzaFastAPIClient(
        StanzaConfiguration(
            api_key="MOCK_API_KEY",
            service_name="fastapi-mock",
            service_release="0.0.1",
            environment="dev",
            hub_address="localhost",
        )
    )


@pytest.fixture(scope="module")
def fastapi_app():
    return FastAPI()


@pytest.fixture(scope="module")
def fastapi_testclient(fastapi_app):
    return TestClient(fastapi_app)


@pytest.mark.asyncio
async def test_async_wrapper_without_request(
    fastapi_client, fastapi_guard, fastapi_app, fastapi_testclient
):
    @fastapi_app.get("/")
    @fastapi_client.stanza_guard("MockGuard")
    async def mock_endpoint(request: Request):
        return {"mock": "response"}

    response = fastapi_testclient.get("/")
    assert response.status_code == 200
    assert response.json() == {"mock": "response"}

    fastapi_guard.__aenter__.assert_called_once()
    fastapi_guard.__aexit__.assert_called_once()


@pytest.mark.asyncio
async def test_async_wrapper_with_request(
    fastapi_client, fastapi_guard, fastapi_app, fastapi_testclient
):
    @fastapi_app.get("/")
    @fastapi_client.stanza_guard("MockGuard")
    async def mock_endpoint():
        return {"mock": "response"}

    response = fastapi_testclient.get("/")
    assert response.status_code == 200
    assert response.json() == {"mock": "response"}

    fastapi_guard.__aenter__.assert_called_once()
    fastapi_guard.__aexit__.assert_called_once()



# import os
from typing import Dict, Optional

from opentelemetry.exporter.otlp.proto.grpc.trace_exporter import OTLPSpanExporter
from opentelemetry.sdk.resources import Resource
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import (
    BatchSpanProcessor,
    ConsoleSpanExporter,
    SpanExporter,
)
from opentelemetry.sdk.trace.sampling import TraceIdRatioBased
from opentelemetry.trace import Tracer

# Constants per Stanza SDK spec
TRACE_EXPORT_INTERVAL = 10_000  # 10,000 milliseconds = 10 seconds
TRACE_EXPORT_TIMEOUT = 30_000  # 30,000 milliseconds = 30 seconds


class StanzaTracerProvider:
    @property
    def provider(self):
        return self.__provider

    def __init__(
        self,
        debug: bool,
        insecure: bool,
        resource: Resource,
        sampler: TraceIdRatioBased,
        endpoint: str,
        headers: Dict[str, str],
    ):
        if debug:
            exporter = self.__debug_span_exporter()
        else:
            exporter = self.__grpc_span_exporter(endpoint, insecure, headers)
        self.__provider = TracerProvider(resource=resource, sampler=sampler)
        self.__provider.add_span_processor(
            BatchSpanProcessor(
                span_exporter=exporter,
                schedule_delay_millis=TRACE_EXPORT_INTERVAL,
                export_timeout_millis=TRACE_EXPORT_TIMEOUT,
            )
        )

    def __debug_span_exporter(self) -> SpanExporter:
        return ConsoleSpanExporter()

    def __grpc_span_exporter(
        self, endpoint: str, insecure: bool, headers: Dict[str, str]
    ) -> SpanExporter:
        return OTLPSpanExporter(
            endpoint=endpoint,
            insecure=insecure,
            # credentials=  # TODO: allow passing credentials
            headers=headers,
            # timeout= # TODO: how is this timeout different from export_timeout_millis?
        )

    def get_tracer(
        self, name: str, version: str, schema_url: Optional[str] = None
    ) -> Tracer:
        return self.__provider.get_tracer(
            instrumenting_module_name=name,
            instrumenting_library_version=version,
            schema_url=schema_url,
        )
